;;;
;;; Copyright 2006, 2007 by Brian Dominy <brian@oddchange.com>
;;; 
;;; This file is part of GCC.
;;; GCC is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;; 
;;; GCC is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; You should have received a copy of the GNU General Public License
;;; along with GCC; if not, write to the Free Software
;;; Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


	;;; Declare external for main() 
	.globl _main

	;;; I/O registers that have meaning within CoCo.
#define COCO_RAMROM_MODE 0xFFDE
#define COCO_ALLRAM_MODE 0xFFDF
#define COCO_PAGE0 0xFFD4
#define COCO_PAGE1 0xFFD5

/* SAM M1 and M0 adjusts the memory size */
#define SAM_M0_CLR 0xffda
#define SAM_M0_SET 0xffdb
#define SAM_M1_CLR 0xffdc
#define SAM_M1_SET 0xffdd

#define BASIC_WARMSTART_FLAG 0x0071
#define BASIC_START 0xA027

	;;; Addresses of other things of interest
#define __STACK_BASE 0x6000
#define __STACK_LEN  0x800


	;;; Provide default objects for all sections
	.area .bss
	.blkb 1

	.area .data
	.byte 0

	.area .ctors
	.word 0

	.area .dtors
	.word 0


	;;;
	;;; __exit : Exit point from the program
	;;; For simulation, this writes to a special I/O register that
	;;; the simulator interprets as end-of-program.
	;;;
	.area	.text
 	.globl __exit
__exit:
	;; Go back to ROM/RAM mode
	sta COCO_RAMROM_MODE
	clr BASIC_WARMSTART_FLAG
	jmp	BASIC_START

	bra	__exit


	;;;
	;;; __start : Entry point to the program
	;;;
	.area	.text
	.globl __start
__start:
	;; Turn off interrupts
	orcc #(0x10|0x40)
	
	;; Setup All RAM Mode
	;;sta SAM_M1_SET
	;;sta SAM_M0_CLR
	;;sta COCO_PAGE0
	sta COCO_ALLRAM_MODE
	
	;; Initialize the stack
	lds	#__STACK_BASE + __STACK_LEN - 2
	
	;; Zero the BSS section

	;; If the _DATA section is in ROM, then copy it
	;; into RAM so that it is modifiable

	;; Initialize the heap

	;; Call any "initializer" functions
	ldu	#s_.ctors
__ctors_loop:
	ldy	,u++
	cmpy	#0
	beq	__ctors_done
	jsr	,y
	bra	__ctors_loop
__ctors_done:

	;; Enable interrupts on the simulator
	;; Not just yet...
	;; andcc	#~(0x10|0x40)

	;; Set up the environment

	;; Set up argc/argv arrays

	;; Call the main function.  The exit code will
	;; be returned in the X register.
	jsr	_main

	;; Call any finalizer functions
	ldu	#s_.dtors
__dtors_loop:
	ldy	,u++
	cmpy	#0
	beq	__dtors_done
	jsr	,y
	bra	__dtors_loop
__dtors_done:

	;; If main returns, then invoke _exit() to stop the program
	jmp	_exit



	;;;
	;;; __crt0_vector : Default handler for interrupts
	;;;
	.area	.text
___crt0_vector:
	;; The default behavior is to simply ignore all
	;; non-reset interrupts.
	rti


	;;;
	;;; vector : The interrupt vector table
	;;; The linker will ensure that this gets loaded at address 0xFFF0.
	;;;
	.area vector
vectors:
	.word ___crt0_vector
	.word ___crt0_vector
	.word ___crt0_vector
	.word ___crt0_vector
	.word ___crt0_vector
	.word ___crt0_vector
	.word ___crt0_vector
	.word __start

	.end __start
