/* libgcc routines for m6809
   Copyright (C) 2006 Free Software Foundation, Inc.

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

In addition to the permissions in the GNU General Public License, the
Free Software Foundation gives you unlimited permission to link the
compiled version of this file with other programs, and to distribute
those programs without any restriction coming from the use of this
file.  (The General Public License restrictions do apply in other
respects; for example, they cover modification of the file, and
distribution when not linked into another program.)

This file is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/* As a special exception, if you link this library with other files,
   some of which are compiled with GCC, to produce an executable,
   this library does not by itself cause the resulting executable
   to be covered by the GNU General Public License.
   This exception does not however invalidate any other reasons why
   the executable file might be covered by the GNU General Public License.  */


#define SIGFPE jmp _abort

#ifdef L_mulhi3
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ___mulhi3 - signed/unsigned multiply
;;; Called by GCC to implement 16x16 multiplication
;;; Arguments: Two 16-bit values, one in stack, one in X.
;;; Result: 16-bit result in X
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	.area .text
	.globl _mulhi3
_mulhi3:
	pshs	x
	lda   5,s   ; left msb * right lsb * 256
	ldb   ,s
	mul
	tfr   b,a
	clrb
	tfr   d,x
	ldb   1,s   ; left lsb * right msb * 256
	lda   4,s
	mul
	tfr   b,a
	clrb
	leax  d,x
	ldb   1,s   ; left lsb * right lsb
	lda   5,s
	mul
	leax  d,x
	puls	d,pc  ; kill D to remove initial push
#endif


#ifdef L_mulhi3_old
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ___mulhi3 - signed/unsigned multiply
;;; Called by GCC to implement 16x16 multiplication
;;; Arguments: Two 16-bit values, both on stack
;;; Result: 16-bit result in D, stack not popped
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	.area .text
	.globl _mulhi3
_mulhi3:
	pshs	x
	ldd   5,s   ; left msb * right lsb * 256
	mul
	tfr   b,a
	clrb
	tfr   d,x
	ldb   7,s   ; left lsb * right msb * 256
	lda   4,s
	mul
	tfr   b,a
	clrb
	leax  d,x
	ldb   7,s   ; left lsb * right lsb
	lda   5,s
	mul
	leax  d,x
	tfr   x,d
	puls	x,pc
#endif


#ifdef L_divhi3
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ___divhi3 - signed division
;;; Arguments: Dividend in X, divisor on the stack
;;; Returns result in X.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	.area .text
	.globl _divhi3
_divhi3:
	ldd	2,s
	bne	do_div		; check dividend
	SIGFPE
do_div:
	pshs	x
	jsr	_seuclid
	puls	x,pc
#endif


#ifdef L_modhi3
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ___modhi3 - signed modulo
;;; Arguments: Dividend in X, divisor on the stack
;;; Returns result in X.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	.area .text
	.globl _modhi3
_modhi3:
	ldd	2,s
	bne	do_mod		; check dividend
	SIGFPE
do_mod:
	pshs	x
	jsr	_seuclid
	leas	2,s
	tfr	d,x
	rts
#endif



#ifdef L_udivhi3
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ___udivhi3 - unsigned division
;;; Arguments: Dividend in X, divisor on the stack
;;; Returns result in X.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	.area .text
	.globl _udivhi3
_udivhi3:
	ldd	2,s
	bne	do_udiv		; check dividend
	SIGFPE
do_udiv:
	pshs	x
	jsr	_euclid
	puls	x,pc
#endif


#ifdef L_umodhi3
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ___umodhi3 - unsigned modulo
;;; Arguments: Dividend in X, divisor on the stack
;;; Returns result in X.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	.area .text
	.globl _umodhi3
_umodhi3:
	ldd	2,s
	bne	do_umod		; check dividend
	SIGFPE
do_umod:
	pshs	x
	jsr	_euclid
	leas	2,s
	tfr	d,x
	rts
#endif


#ifdef L_euclid
;	unsigned euclidean division
;	calling: (left / right)
;		push left
;		ldd right
;		jsr _euclid
;	quotient on the stack (left)
;	modulus in d

	.area	.text
	.globl	_euclid
	left=5
	right=1			; word
	count=0			; byte
	CARRY=1			; alias
_euclid:
	leas	-3,s		; 2 local variables
	clr	count,s		; prescale divisor
	inc	count,s
	tsta
presc:
	bmi	presc_done
	inc	count,s
	aslb
	rola
	bra	presc
presc_done:
	std	right,s
	ldd	left,s
	clr	left,s		; quotient = 0
	clr	left+1,s
mod1:
	subd	right,s		; check subtract
	bcc	mod2
	addd	right,s
	andcc	#~CARRY
	bra	mod3
mod2:
	orcc	#CARRY
mod3:
	rol	left+1,s	; roll in carry
	rol	left,s
	lsr	right,s
	ror	right+1,s
	dec	count,s
	bne	mod1
	leas	3,s
	rts
#endif

#ifdef L_seuclid
;	signed euclidean division
;	calling: (left / right)
;		push left
;		ldd right
;		jsr _seuclid
;	quotient on the stack (left)
;	modulus in d
	.area	.text
	.globl	_seuclid
	left=6
	right=2
	quot_sign=1
	mod_sign=0
_seuclid:
	leas	-4,s		; 3 local variables
	std	right,s
	clr	mod_sign,s
	clr	quot_sign,s
	ldd	left,s
	bge	mod_abs
	inc	mod_sign,s	; sign(mod) = sign(left)
	inc	quot_sign,s
	bsr	negd		; abs(left) -> D
mod_abs:
	pshs	b,a		; push abs(left)
	ldd	right+2,s	; all references shifted by 2
	bge	quot_abs
	dec	quot_sign+2,s	; sign(quot) = sign(left) XOR sign(right)
	bsr	negd		; abs(right) -> D
quot_abs:
	jsr	_euclid		; call (unsigned) euclidean division
	std	right+2,s
	puls	a,b		; quot -> D
	tst	quot_sign,s	; all references no longer shifted
	beq	quot_done
	bsr	negd
quot_done:
	std	left,s		; quot -> left
	ldd	right,s
	tst	mod_sign,s
	beq	mod_done
	bsr	negd
mod_done:
	leas	4,s		; destroy stack frame
	rts

negd:				; self-explanatory !
	nega
	negb
	sbca	#0
	rts
#endif



#ifdef L_pending_addsi3
_addsi3:
	rts
#endif /* L_pending_addsi3 */



