/*-------------------------------------------------------------------
	FILE: m6809.c
-------------------------------------------------------------------*/
/* Subroutines for insn-output.c for MC6809.
   Copyright (C) 1989 Free Software Foundation, Inc.

 MC6809 Version by Tom Jones (jones@sal.wisc.edu)
 Space Astronomy Laboratory
 University of Wisconsin at Madison

 minor changes to adapt it to gcc-2.5.8 by Matthias Doerfel
 ( msdoerfe@informatik.uni-erlangen.de )
 also added #pragma interrupt (inspired by gcc-6811)

 minor changes to adapt it to gcc-2.8.0 by Eric Botcazou
 (ebotcazou@multimania.com)

 minor changes to adapt it to gcc-2.95.3 by Eric Botcazou
 (ebotcazou@multimania.com)

 major cleanup, improvements, and upgrade to gcc 3.4 by Brian Dominy
 (brian@oddchange.com)

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
 * This GCC machine configuration has been designed to be portable to
 * a number of different gcc versions at once.  TARGET_GCC_VERSION
 * is used to determine the exact version at compile time.  There are
 * many differences that must be handled by testing this value.
 */

#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "tree.h"
#include "rtl.h"
#include "regs.h"
#include "flags.h"
#include "hard-reg-set.h"
#include "real.h"
#include "tree.h"
#include "insn-config.h"
#include "conditions.h"
#include "insn-flags.h"
#include "output.h"
#include "insn-attr.h"
#include "function.h"
#include "target.h"
#include "target-def.h"
#include "expr.h"
#include "recog.h"
#include "cpplib.h"
#include "c-pragma.h"
#include "toplev.h"
#include "tm_p.h"


/** Define GCC_SUBVERSION_TAG to a string value to identify
 * further this particular version of the compiler.  The tag
 * value is displayed in assembler listings. */
#define C_STRING(x) C_STR(x)
#define C_STR(x) #x
static const char subversion_tag[] =
#ifdef GCC_SUBVERSION_TAG
	C_STRING(GCC_SUBVERSION_TAG)
#else
	"[no tag]"
#endif
	;

/* macro to return TRUE if length of operand mode is one byte */
#define BYTE_MODE(X) ((GET_MODE_SIZE (GET_MODE (X))) == 1)

/* REAL_REG_P(x) is a true if the rtx 'x' represents a real CPU
register and not a fake one that is emulated in software. */
#define REAL_REG_P(x) (REG_P(x) && !M_REG_P(x))

/*-------------------------------------------------------------------
    Target hooks, moved from target.h
-------------------------------------------------------------------*/
static void m6809_encode_section_info PARAMS ((tree, int ));

#undef TARGET_ENCODE_SECTION_INFO
#define TARGET_ENCODE_SECTION_INFO m6809_encode_section_info

#undef TARGET_ASM_FILE_START
#define TARGET_ASM_FILE_START m6809_asm_file_start

#undef TARGET_ASM_ALIGNED_HI_OP
#define TARGET_ASM_ALIGNED_HI_OP "\t.word\t"

#undef TARGET_ASM_ALIGNED_SI_OP
#define TARGET_ASM_ALIGNED_SI_OP NULL

#undef TARGET_ASM_UNALIGNED_HI_OP
#define TARGET_ASM_UNALIGNED_HI_OP "\t.word\t"

#undef TARGET_ASM_UNALIGNED_SI_OP
#define TARGET_ASM_UNALIGNED_SI_OP NULL

#undef TARGET_RTX_COSTS
#define TARGET_RTX_COSTS m6809_rtx_costs

#undef TARGET_ATTRIBUTE_TABLE
#define TARGET_ATTRIBUTE_TABLE m6809_attribute_table

#undef TARGET_INIT_BUILTINS
#define TARGET_INIT_BUILTINS m6809_init_builtins

#undef TARGET_EXPAND_BUILTIN
#define TARGET_EXPAND_BUILTIN m6809_expand_builtin


/* External variables used */
extern int reload_completed;   /* set in toplev.c */
extern FILE *asm_out_file;

static int last_mem_size;   /* operand size (bytes) */

/* True if the section was recently changed and another .area
 * directive needs to be output before emitting the next label. */
int section_changed = 0;

/* Section names.  The defaults here are used until an
 * __attribute__((section)) is seen that changes it. */
char code_section_op[128] = "\t.area .text";
char data_section_op[128] = "\t.area .data";
char bss_section_op[128] = "\t.area .bss";

const char *code_section_ptr = 0;
const char *data_section_ptr = 0;
const char *bss_section_ptr = 0;

/* TRUE if the direct mode prefix might be valid in this context.
 * This is set by 'print_address' prior to calling output_addr_const,
 * which performs into 'print_direct_prefix' to do the final checks. */
static int check_direct_prefix_flag;

/* Nonzero if an address is being printed in a context which does not
 * permit any PIC modifications to the address */
static int pic_ok_for_addr_p = 1;

/* Current code page.  This supports machines which can do bank
 * switching to allow for more than 64KB of code/data. */
char far_code_page[64];
const char *far_code_page_option = 0;

/* Current bank name */
static char current_bank_name[8] = "-1";

/* Default bank name */
static char default_code_bank_name[8] = "-1";
const char *code_bank_option = 0;

/* Direct memory reserved as soft registers */
const char *m6809_soft_reg_count = 0;
unsigned int m6809_soft_regs = 0;

/* ABI version */
const char *m6809_abi_version_ptr = 0;
unsigned int m6809_abi_version = M6809_ABI_VERSION_REGS;


/**
 * Called after options have been parsed.
 * If overrides have been specified on the command-line, then
 * these values are copied into the main storage variables.
 */
void
m6809_override_options (void)
{
	/* -fpic is not supported yet */
	if (flag_pic)
		warning ("-f%s for 6809 not stable", (flag_pic > 1) ? "PIC" : "pic");

	/* Handle -mfar-code-page */
	if (far_code_page_option == 0)
		far_code_page_option = "__default_code_page";
	strcpy (far_code_page, far_code_page_option);

	/* Handle -mcode-section, -mdata-section, and -mbss-section */
	if (code_section_ptr != 0)
		sprintf (code_section_op, "\t.area %s", code_section_ptr);
	if (data_section_ptr != 0)
		sprintf (data_section_op, "\t.area %s", data_section_ptr);
	if (bss_section_ptr != 0)
		sprintf (bss_section_op, "\t.area %s", bss_section_ptr);

	/* Handle -mcode-bank */
	if (code_bank_option != 0)
		sprintf (default_code_bank_name, "%s", code_bank_option);

	/* Handle -mabi-version or -mno-reg-args */
	if (m6809_abi_version_ptr != 0)
	{
		if (!strcmp (m6809_abi_version_ptr, "stack"))
			m6809_abi_version = M6809_ABI_VERSION_STACK;
		else if (!strcmp (m6809_abi_version_ptr, "regs"))
			m6809_abi_version = M6809_ABI_VERSION_REGS;
		else if (!strcmp (m6809_abi_version_ptr, "bx"))
			m6809_abi_version = M6809_ABI_VERSION_BX;
		else if (!strcmp (m6809_abi_version_ptr, "latest"))
			m6809_abi_version = M6809_ABI_VERSION_LATEST;
		else
			m6809_abi_version = atoi (m6809_abi_version_ptr);
	}

	/* The older -mno-reg-args option is deprecated, and treated
	as -mabi=stack. */
	if (!TARGET_REG_ARGS)
   {
      warning ("-mno-reg-args deprecated; use -mabi=stack instead.");
      m6809_abi_version = M6809_ABI_VERSION_STACK;
   }
}


/**
 * Output prefix that directs the assembler to use a direct-mode
 * instruction if globally enabled, address is a symbol, and symbol
 * has been marked as in direct page.  Also, never do this if
 * using the indirect mode. */
void
print_direct_prefix (FILE * file, rtx addr)
{
	if (TARGET_DIRECT &&
       (GET_CODE (addr) == SYMBOL_REF) && 
       SYMBOL_REF_FLAG (addr) &&
       check_direct_prefix_flag)
   {
      putc ('*', file);
   }
}


/** Predicate like 'register_operand', but which does not allow
 * SUBREGs. */
int
whole_register_operand (rtx op, enum machine_mode mode)
{
	return (register_operand (op, mode) && (GET_CODE(op) != SUBREG));
}


/** Likewise, a replacement for 'general_operand' which excludes
 * SUBREGs. */
int
whole_general_operand (rtx op, enum machine_mode mode)
{
	return (general_operand (op, mode) && (GET_CODE(op) != SUBREG));
}


/** A predicate for immediate shift counts that can be done
 * reasonably fast. */
int
shift_count_operand (rtx op, enum machine_mode mode ATTRIBUTE_UNUSED)
{
	return GET_CODE (op) == CONST_INT
		&& ((INTVAL (op) == 1) || (INTVAL (op) == 8));
}


int
symbolic_operand (rtx op, enum machine_mode mode ATTRIBUTE_UNUSED)
{
	return GET_CODE (op) == SYMBOL_REF;
}


/** A predicate that matches any bitwise logical operator.  This
 * allows for a single RTL pattern to be used for multiple operations. */
int
logical_bit_operator (rtx op, enum machine_mode mode)
{
	return ((mode == VOIDmode || GET_MODE (op) == mode)
		&& (GET_CODE (op) == AND || GET_CODE(op) == IOR || GET_CODE(op) == XOR));
}


/** A predicate that matches any shift or rotate operator.  This
 * allows for a single RTL pattern to be used for multiple operations. */
int
shift_rotate_operator (rtx op, enum machine_mode mode)
{
	return ((mode == VOIDmode || GET_MODE (op) == mode)
		&& (GET_CODE (op) == ASHIFT
			|| GET_CODE(op) == ASHIFTRT
			|| GET_CODE(op) == LSHIFTRT
			|| GET_CODE(op) == ROTATE
			|| GET_CODE(op) == ROTATERT));
}


/** Prints an operand (that is not an address) in assembly from RTL. */
void
print_operand (FILE * file, rtx x, int code)
{
	if (REG_P (x)) {
		/* gcc currently allocates the entire 16-bit 'd' register
		 * even when it only needs an 8-bit value.  So here it
		 * is tricked into printing only the lower 8-bit 'b'
		 * register into the assembly output.
		 *
		 * Eventually gcc should be modified to allocate a/b
		 * independently and this hack can be removed.
		 *
		 * Occasionally, we may want to do an operation using
		 * the 'a' register instead of 'b'; use the 'A' code
		 * to specify that.
		 */
		if (code == 'A')
			fputs ("a", file);
		else if ((BYTE_MODE (x)) && (REGNO (x) == HARD_D_REGNUM))
			fputs ("b", file);
		else
			fputs (reg_names[REGNO (x)], file);
	}

	else if (MEM_P (x)) {
		last_mem_size = GET_MODE_SIZE (GET_MODE (x));
		if (code == 'L') {	/* LSH of word address */
			if (GET_CODE (XEXP (x, 0)) == MEM)
			{
				/* Offseting an indirect addressing mode is not supported */
				error ("expression too complex for 6809 (offset indirect mode)");
				debug_rtx (x);
			}
			else
				x = adjust_address (x, QImode, 1);
		}
		else if (code == 'M') { /* MSH of word address */
			if (GET_CODE (XEXP (x, 0)) == MEM)
			{
				/* Offseting an indirect addressing mode is not supported */
				error ("expression too complex for 6809 (offset indirect mode)");
				debug_rtx (x);
			}
			else
				x = adjust_address (x, QImode, 0);
		}

		pic_ok_for_addr_p = (code != 'C');
		output_address (XEXP (x, 0));
	}

	else if (GET_CODE (x) == CONST_DOUBLE && GET_MODE (x) != DImode) {
		union { double d; int i[2]; } u;
		u.i[0] = CONST_DOUBLE_LOW (x);
		u.i[1] = CONST_DOUBLE_HIGH (x);
		fprintf (file, "#%#9.9g", u.d);
	}

	else if (code == 'R') {
		fprintf (file, "%s", 
			m6809_get_regs_printable (INTVAL (x)));
	}

	else {
		if (code == 'L') {	/* LSH of word address */
			x = gen_rtx_CONST_INT (VOIDmode, (INTVAL(x) & 0xff));
		}
		else if (code == 'M') {	/* MSH of word address */
			x = gen_rtx_CONST_INT (VOIDmode, ((INTVAL(x) >> 8) & 0xff));
		}

		putc ('#', file);
		output_addr_const (file, x);
	}
}


/** Prints an address operand to assembler from its RTL representation. */
void
print_operand_address (file, addr)
	FILE *file;
	register rtx addr;
{
	register rtx base = 0;
	register rtx offset = 0;
	int regno;
	int indirect_flag = 0;

	check_direct_prefix_flag = 0;

	/*** check for indirect addressing ***/
	if (MEM_P (addr)) {
		last_mem_size = GET_MODE_SIZE (GET_MODE (addr));
		addr = XEXP (addr, 0);
		if (pic_ok_for_addr_p)
		{
			indirect_flag = 1;
			fprintf (file, "[");
		}
	}


	switch (GET_CODE (addr)) {
		case REG:
			regno = REGNO (addr);
			fprintf (file, ",%s", reg_names[regno]);
			break;

		case PRE_DEC:
			regno = REGNO (XEXP (addr, 0));
			fputs (((last_mem_size == 1) ? ",-" : ",--"), file);
			fprintf (file, "%s", reg_names[regno]);
			break;

		case POST_INC:
			regno = REGNO (XEXP (addr, 0));
			fprintf (file, ",%s", reg_names[regno]);
			fputs (((last_mem_size == 1) ? "+" : "++"), file);
			break;

		case PLUS:
			base = XEXP (addr, 0);
			if (MEM_P (base))
				base = XEXP (base, 0);

			offset = XEXP (addr, 1);
			if (MEM_P (offset))
				offset = XEXP (offset, 0);

			if ((CONSTANT_ADDRESS_P (base)) && (CONSTANT_ADDRESS_P (offset))) {
				if (!indirect_flag)
					check_direct_prefix_flag = 1;
				output_addr_const (file, base);
				check_direct_prefix_flag = 0;
				fputs ("+", file);
				output_addr_const (file, offset);
			}

			else if ((CONSTANT_ADDRESS_P (base)) && (A_REG_P (offset))) {
				output_addr_const (file, base);
				fprintf (file, ",%s", reg_names[REGNO (offset)]);
			}

			else if ((CONSTANT_ADDRESS_P (offset)) && (A_REG_P (base))) {
				output_addr_const (file, offset);
				fprintf (file, ",%s", reg_names[REGNO (base)]);
			}

			/*** accumulator offset ***/
			else if (((D_REG_P (offset)) || (Q_REG_P (offset)))
			&& (A_REG_P (base))) {
				fprintf (file, "%s,%s",
				reg_names[REGNO (offset)], reg_names[REGNO (base)]);
			}

			else if (((D_REG_P (base)) || (Q_REG_P (base)))
			&& (A_REG_P (offset))) {
				fprintf (file, "%s,%s",
				reg_names[REGNO (base)], reg_names[REGNO (offset)]);
			}

			else if (GET_CODE (base) == PRE_DEC) {
				regno = REGNO (XEXP (base, 0));
				fputs (((last_mem_size == 1) ? ",-" : ",--"), file);
				fprintf (file, "%s", reg_names[regno]);
			}

			else
				abort ();

			break;

   default:
		/* Set this global before calling output_addr_const() */
		if (!indirect_flag)
			check_direct_prefix_flag = 1;

		/* When printing a SYMBOL_REF in PIC mode, do not print the leading
		 * '#' and follow it by ',pcr' to enable relative addressing. */
		if (flag_pic && pic_ok_for_addr_p && GET_CODE (addr) == SYMBOL_REF)
		{
			ASM_OUTPUT_SYMBOL_REF (file, addr);
			fputs (",pcr", file);
			pic_ok_for_addr_p = 1;
		}
		else
		{
      	output_addr_const (file, addr);
		}

		check_direct_prefix_flag = 0;
      break;
	}

	if (indirect_flag)
		fprintf (file, "]");
}

/*-------------------------------------------------------------------
    Update the CC Status
---------------------------------------------------------------------
   Set the cc_status for the results of an insn whose pattern is EXP.
   We assume that jumps don't affect the condition codes.
   All else, clobbers the condition codes, by assumption.

   We assume that ALL add, minus, etc. instructions effect the condition
   codes.
-------------------------------------------------------------------*/
void
notice_update_cc (rtx exp, rtx insn ATTRIBUTE_UNUSED)
{
    /*** recognize SET insn's ***/
    if (GET_CODE (exp) == SET) {
	/* Jumps do not alter the cc's.  */
	if (SET_DEST (exp) == pc_rtx)
	    return;

	/* Moving one register into another register (tfr):
	Doesn't alter the cc's.  */
	if (REG_P (SET_DEST (exp)) && (REG_P (SET_SRC (exp)))) {
	    return;
	}

	/* Moving memory into a register (load): Sets cc's. */
	if (REG_P (SET_DEST (exp)) && MEM_P (SET_SRC (exp))) {
	    cc_status.value1 = SET_SRC (exp);
	    cc_status.value2 = SET_DEST (exp);
	    return;
	}

	/* Moving register into memory (store): Sets cc's. */
	if (MEM_P (SET_DEST (exp)) && REG_P (SET_SRC (exp))) {
	    cc_status.value1 = SET_SRC (exp);
	    cc_status.value2 = SET_DEST (exp);
	    return;
	}

	/* Function calls clobber the cc's.  */
	else if (GET_CODE (SET_SRC (exp)) == CALL) {
	    CC_STATUS_INIT;
	    return;
	}

	/* Tests and compares set the cc's in predictable ways.  */
	else if (SET_DEST (exp) == cc0_rtx) {
	    cc_status.flags = 0;
	    cc_status.value1 = SET_SRC (exp);
	    cc_status.value2 = SET_DEST (exp);
	    return;
	}

	/* Certain instructions effect the condition codes. */
	else if (GET_MODE (SET_SRC (exp)) == SImode
	    || GET_MODE (SET_SRC (exp)) == HImode) {
		switch (GET_CODE (SET_SRC (exp))) {
		case PLUS: case MINUS: case NEG:
		case AND: case IOR: case XOR:
		if (A_REG_P (SET_DEST (exp))) {
		    CC_STATUS_INIT;
		}
		else {
		    cc_status.value1 = SET_SRC (exp);
		    cc_status.value2 = SET_DEST (exp);
		}
		    break;

		default:
		    CC_STATUS_INIT;
		}
	    return;
	}
    }
    else if (GET_CODE (exp) == PARALLEL
	&& GET_CODE (XVECEXP (exp, 0, 0)) == SET) {
	    if (SET_DEST (XVECEXP (exp, 0, 0)) == pc_rtx)
		return;
	if (SET_DEST (XVECEXP (exp, 0, 0)) == cc0_rtx) {
	    CC_STATUS_INIT;
	    cc_status.value1 = SET_SRC (XVECEXP (exp, 0, 0));
	    return;
	}
    }

    /* Default action for any other instructions is to assume
     * nothing about the condition code status */
    CC_STATUS_INIT;
}


/* strip path from the filename
 *  for AS6809 assembler.
 */
static char *
my_basename (const char *filename)
{
   char *p = filename + strlen(filename) - 1;

   while ((p >= filename) && (*p != '/'))
      p--;

   return p+1;
}




/** Return a bitarray of the hard registers which are used by a function. */
unsigned int
m6809_get_live_regs (void)
{
	unsigned int regs = 0;
	int regno;

	if (frame_pointer_needed)
		regs |= (1 << HARD_FRAME_POINTER_REGNUM);

	for (regno = HARD_X_REGNUM; regno <= HARD_U_REGNUM; regno++)
		if (regs_ever_live[regno] && ! call_used_regs[regno])
			regs |= (1 << regno);

	return regs;
}


/** Return a printable version of a list of hard registers, suitable
 * for use in a PSHx or PULx insn. */
const char *
m6809_get_regs_printable (unsigned int regs)
{
	static char list[64];
	char *listp = list;
	unsigned int regno;

	for (regno=0; regno < FIRST_PSEUDO_REGISTER; regno++)
		if ((regs & (1 << regno)) && !S_REGNO_P (regno))
			listp += sprintf (listp,
				(listp == list) ? "%s" : ",%s", reg_names[regno]);

	return list;
}


/** Return the total number of bytes covered by a set of registers. */
unsigned int
m6809_get_regs_size (unsigned int regs)
{
	unsigned int regno;
	unsigned int size = 0;

	for (regno=0; regno < FIRST_PSEUDO_REGISTER; regno++)
	{
		/* Only count register in the given register set */
		if (REGSET_CONTAINS_P (regno, regs))
		{
			/* Add 1 or 2 byte, depending on the size of the register.
			 * Since 'D' may be in both sets, check for WORD_REGSET first. */
			if (REGSET_CONTAINS_P(regno, WORD_REGSET))
				size += 2;
			else if (REGSET_CONTAINS_P(regno, BYTE_REGSET))
				size++;
		}
	}
	return size;
}


/* Given the target of call instruction in X,
 * return the tree node that contains the function declaration for
 * that target.
 *
 * If the rtx or the tree do not appear valid for any reason,
 * then return NULL_TREE.
 */
static tree call_target_decl (rtx x)
{
   tree decl;

	/* Make sure the target is really a MEM. */
	if (!x || !MEM_P (x))
		return NULL_TREE;

	/* Make sure the address is a SYMBOL_REF. */
	x = XEXP (x, 0);
	if (!x || (GET_CODE (x) != SYMBOL_REF))
		return NULL_TREE;

	/* Get the declaration of this symbol */
	decl = SYMBOL_REF_DECL (x);

	/* Make sure the declaration is really a function. */
	if (!decl || (TREE_CODE(decl) != FUNCTION_DECL))
		return NULL_TREE;

   return decl;
}


/** Returns nonzero if a function, whose declaration is in DECL,
 * was declared to have the attribute given by ATTR_NAME. */
int
m6809_function_has_type_attr_p (tree decl, const char *attr_name)
{
	tree type;

	type = TREE_TYPE (decl);
	return lookup_attribute (attr_name, TYPE_ATTRIBUTES (type)) != NULL;
}



/** Returns nonzero if the current function was declared to have the
 * attribute given by ATTR_NAME. */
int
m6809_current_function_has_type_attr_p (const char *attr_name)
{
	return m6809_function_has_type_attr_p (current_function_decl, attr_name);
}


/** Return nonzero if the current function has no return value. */
int
m6809_current_function_is_void (void)
{
	return 0; /* TODO */
}


/** Get the value of a declaration's 'bank', as set by the 'bank'
 * attribute.  If no bank was declared, it returns NULL by default. */
const char *
m6809_get_decl_bank (tree decl)
{
	tree attr;

	/* Lookup the 'bank' attribute.  If it does not exist, then
	 * return NULL */
	attr = lookup_attribute ("bank", DECL_ATTRIBUTES (decl));
	if (attr == NULL_TREE)
		return NULL;

	/* Make sure it has a value assigned to it */
	attr = TREE_VALUE (attr);
	if (attr == NULL_TREE)
	{
		warning ("banked function did not declare a bank number");
		return NULL;
	}

	/* Return the bank name */
	attr = TREE_VALUE (attr);
	return TREE_STRING_POINTER (attr);
}


void
m6809_declare_function_name (FILE *asm_out_file, const char *name, tree decl)
{
	/* Check the function declaration for special properties.
	 *
	 * If the function was declare with __attribute__((bank)), output
	 * assembler definitions to force the function to go into the named
	 * bank.
	 */
	const char *bank_name = m6809_get_decl_bank (decl);
	if (bank_name != NULL)
	{
		/* Declare __self_bank as a local assembler value that denotes
		 * which bank the current function is in.  This is required only
		 * when the bank actually changes. */
		if (strcmp (bank_name, current_bank_name))
		{
			fprintf (asm_out_file, "__self_bank\t.equ %s\n", bank_name);
			strcpy (current_bank_name, bank_name);
		}

		/* Declare a global assembler value that denotes which bank the
		 * named function is in. */
		fprintf (asm_out_file, "__%s_bank\t.gblequ %s\n", name, bank_name);

		/* Force the current function into a new area */
		fprintf (asm_out_file, "\t.bank bank_%s (FSFX=_%s)\n",
			bank_name, bank_name);
		fprintf (asm_out_file, "\t.area bank_%s (BANK=bank_%s)\n",
			bank_name, bank_name);
	}

	/* Emit the label for the function's name */
	ASM_OUTPUT_LABEL (asm_out_file, name);
}


int
m6809_get_actual_frame_size (int register_flag)
{
   unsigned int frame_size = 0;

   /*
    * Compute the number of bytes of stack size needed to
    * be saved/restored.  This is the sum of:
    *    1. The outgoing argument size, if enabled
    *    2. The number of bytes for local variables
    *    3. The space for saving/restoring used registers.
	 *    Include this only if register_flag is nonzero.
    */
	if (ACCUMULATE_OUTGOING_ARGS)
		frame_size += current_function_outgoing_args_size;

   frame_size += get_frame_size ();

	if (register_flag)
   	frame_size += m6809_get_regs_size (m6809_get_live_regs ());
	
	return (frame_size);
}


/*
 * Handle pragmas.  Note that only the last branch pragma seen in the 
 * source has any affect on code generation.  
 */
#define BAD_PRAGMA(msgid, arg) \
	do { warning (msgid, arg); return -1; } while (0)

static int
pragma_parse (const char *name, tree *sect)
{
  tree s, x;

  if (c_lex (&x) != CPP_OPEN_PAREN)
    BAD_PRAGMA ("missing '(' after '#pragma %s' - ignored", name);

  if (c_lex (&s) != CPP_STRING)
    BAD_PRAGMA ("missing section name in '#pragma %s' - ignored", name);

  if (c_lex (&x) != CPP_CLOSE_PAREN)
    BAD_PRAGMA ("missing ')' for '#pragma %s' - ignored", name);

  if (c_lex (&x) != CPP_EOF)
    warning ("junk at end of '#pragma %s'", name);

  *sect = s;
  return 0;
}


/*
 * Handle #pragma section.
 * This is deprecated; code should use __attribute__(section("name"))
 * instead.
 */
void pragma_section (cpp_reader *pfile ATTRIBUTE_UNUSED)
{
	tree sect;

	if (pragma_parse ("section", &sect))
		return;

	snprintf (code_section_op, 6+TREE_STRING_LENGTH (sect),
		".area\t%s", TREE_STRING_POINTER (sect));
	snprintf (data_section_op, 6+TREE_STRING_LENGTH (sect),
		".area\t%s", TREE_STRING_POINTER (sect));

	/* Mark a flag that sections have changed.  Upon emitting another
	 * declaration, the new .area directive will be written. */
	section_changed++;
}


/**
 * Check a `double' value for validity for a particular machine mode.
 * Called by the CHECK_FLOAT_VALUE() machine-dependent macro.
 */
int
check_float_value (enum machine_mode mode, double *d, int overflow)
{
	if (mode == SFmode) {
		if (*d > 1.7014117331926443e+38) {
			error("magnitude of constant too large for `float'");
			*d = 1.7014117331926443e+38;
		}
		else if (*d < -1.7014117331926443e+38) {
			error("magnitude of constant too large for `float'");
			*d = -1.7014117331926443e+38;
		}
		else if ((*d > 0) && (*d < 2.9387358770557188e-39)) {
			warning("`float' constant truncated to zero");
			*d = 0.0;
		}
		else if ((*d < 0) && (*d > -2.9387358770557188e-39)) {
			warning("`float' constant truncated to zero");
			*d = 0.0;
		}
	}
	return overflow;
}


/** Initialize the state of function call arguments. */
int
m6809_init_cumulative_args (CUMULATIVE_ARGS cum ATTRIBUTE_UNUSED,
     tree fntype,
     rtx libname ATTRIBUTE_UNUSED)
{
	cum = 0;

	if (fntype && TYPE_ARG_TYPES (fntype) != 0 &&
		(TREE_VALUE (tree_last (TYPE_ARG_TYPES (fntype))) != void_type_node))
	{
		/* has variable arguments, cannot use registers */
		cum |= (CUM_X_MASK | CUM_B_MASK | CUM_STACK_ONLY);
	}

	if (m6809_abi_version == M6809_ABI_VERSION_STACK)
	{
		/* cannot use registers ; only use the stack */
		cum |= (CUM_STACK_ONLY | CUM_X_MASK | CUM_B_MASK);
	}

	return cum;
}


/** Declare that the target supports named output sections. */
bool m6809_have_named_section = (bool)1;


/** Write to the assembler file a directive to place
 * subsequent objects to a different section in the
 * object file.  ASxxxx uses the "area" directive for
 * this purpose.  It does not however support generalized
 * alignment, and can only place items on an odd/even
 * boundary. */
void
m6809_asm_named_section (
	const char *name, 
	unsigned int flags ATTRIBUTE_UNUSED,
	unsigned int align)
{
	if (align > 2)
	{
		warning ("alignment greater than 2 not supported");
		align = 2;
	}
	if (align == 2)
		fprintf (asm_out_file, ".even\n");

	fprintf (asm_out_file, "\t.area\t%s\n", name);
}


enum reg_class
m6809_preferred_reload_class (rtx x, enum reg_class regclass)
{
	/* Check cases based on type code of rtx */
	switch (GET_CODE(x))
	{
		case CONST_INT:
		   /* Constants that can fit into 1 byte should be
			 * loaded into a Q_REGS reg */
			if (((unsigned) (INTVAL(x) + 0x80) < 0x100) &&
  				 (regclass > A_REGS))
      		return Q_REGS;

			/* 16-bit constants should be loaded into A_REGS
			 * when possible.  gcc may already require A_REGS
			 * or D_REGS for certain types of instructions.
			 * This case applies mostly to simple copy operations
			 * to/from memory when any register will do, but
			 * it's best to avoid using D register since it is
			 * needed for other things.
			 */
			else if (((unsigned) (INTVAL(x) + 0x80) < 0x10000) &&
  				 (regclass > A_REGS))
      		return A_REGS;
			break;

		case SYMBOL_REF:
		case LABEL_REF:
			/* Addresses should always be loaded into A_REGS */
			if (regclass >= A_REGS)
				return (A_REGS);

		default:
			break;
	}

	/* Check cases based on mode of rtx */
   if ((GET_MODE(x) == QImode) && (regclass != A_REGS))
      return Q_REGS;

	/* Default: return whatever class reload suggested */
   return regclass;
}


/**
 * Check a new declaration for the "section" attribute.
 * If it exists, and the target section is "direct", then mark
 * the declaration (in RTL) to indicate special treatment.
 * When the variable is referenced later, we test for this flag
 * and can emit special asm text to force the assembler to use
 * short instructions.
 */
static void
m6809_encode_section_info (tree decl, int new_decl_p ATTRIBUTE_UNUSED)
{
   tree attr, id;
   const char *name;
   const char *decl_name;

   /* We only care about variable declarations, not functions */
   if (TREE_CODE (decl) != VAR_DECL)
      return;

	/* For debugging purposes only; grab the decl's name */
   decl_name = IDENTIFIER_POINTER (DECL_NAME (decl));

	/* Give up if the decl doesn't have any RTL */
   if (!DECL_RTL (decl))
      return;

	/* See if it has a section attribute */
   attr = lookup_attribute ("section", DECL_ATTRIBUTES (decl));
   if (!attr)
      return;

	/* See if the section attribute has a value */
   id = TREE_VALUE (TREE_VALUE (attr));
   if (!id)
      return;
   name = TREE_STRING_POINTER (id);
   if (!name)
      return;

	/* See if the value is 'direct'.  If so, mark it. */
   if (!strcmp (name, "direct"))
      SYMBOL_REF_FLAG (XEXP (DECL_RTL (decl), 0)) = 1;
}


/**
 * Output code to perform a complex shift, for which there is no
 * direct support in the instruction set.
 *
 * shift1 is an instruction pattern for performing a 1-bit modification.
 * This code wraps that pattern in a loop to perform the shift N times,
 * where N is given by operands[2].
 *
 * To support 16-bit shifts, shift2 can also be provided: it is
 * a second instruction to be included in the loop.  8-bit shift
 * insns will pass NULL here.
 *
 * The insn length of shift1/shift2 is assumed to be 1 byte,
 * which works in all of the cases it is needed so far.
 */
static void
m6809_gen_register_shift (
		rtx *operands, 
		const char *shift1, 
		const char *shift2 )
{
	char beq_pattern[32];
	char bra_pattern[32];

	int shiftlen = (shift1 && shift2) ? 2 : 1;
	int cmplen = (REGNO (operands[2]) == HARD_X_REGNUM) ? 3 : 4;

	int beq_offset = 2 + shiftlen + 2;
	int bra_offset = shiftlen + 2 + cmplen + 2;

	sprintf (beq_pattern, "beq\t.+%d", beq_offset);
	sprintf (bra_pattern, "bra\t.-%d", bra_offset);

	output_asm_insn ("pshs\t%2", operands);
	output_asm_insn ("lea%2\t-1,%2", operands);
   output_asm_insn ("cmp%2\t#-1", operands);
   output_asm_insn (beq_pattern, operands);
	if (shift1)
   	output_asm_insn (shift1, operands);
	if (shift2)
   	output_asm_insn (shift2, operands);
  	output_asm_insn (bra_pattern, operands);
	output_asm_insn ("puls\t%2", operands);
}


/** Generate RTL for the upper 8-bits of a 16-bit constant. */
rtx
gen_rtx_const_high (rtx r)
{
   unsigned char v = (INTVAL (r) >> 8) & 0xFF;
	signed char s = (signed char)v;
   return gen_rtx_CONST_INT (QImode, s);
}


/** Generate RTL for the lower 8-bits of a 16-bit constant. */
rtx
gen_rtx_const_low (rtx r)
{
   unsigned char v = INTVAL (r) & 0xFF;
	signed char s = (signed char)v;
   return gen_rtx_CONST_INT (QImode, s);
}


/** Generate RTL to allocate/free bytes on the stack.
 * CODE is given as MINUS when allocating and PLUS when freeing,
 * to match the semantics of a downward-growing stack.  SIZE
 * is always given as a positive integer.
 */
static rtx
gen_rtx_stack_adjust (enum rtx_code code, int size)
{
	if (size <= 0)
		return NULL_RTX;

	if (code == MINUS)
		size = -size;

	return gen_rtx_SET (Pmode, stack_pointer_rtx, 
		gen_rtx_PLUS (Pmode, stack_pointer_rtx,
   		gen_rtx_CONST_INT (HImode, size)));
}


/** Generate RTL to push/pop a set of registers. */
rtx
gen_rtx_register_pushpop (int op, int regs)
{
	rtx nregs = gen_rtx_CONST_INT (QImode, regs);
	
	if (op == UNSPEC_PUSH_RS)
		return gen_register_push (nregs);
	else
		return gen_register_pop (nregs);
}


/* Given a register set REGS, where the bit positions correspond to
 * hard register numbers, return another bitmask that represents the
 * order in which those registers would be pushed/popped.
 * Registers that are pushed first have higher bit positions.
 * The pop order is just the reverse bitmask.
 * These values are the same as the bitmasks actually used in the
 * machine instructions. */
static unsigned int
register_push_order (int regs)
{
	unsigned int order = 0;

	if (REGSET_CONTAINS_P (HARD_PC_REGNUM, regs))
		order |= 0x80;
	if (REGSET_CONTAINS_P (HARD_U_REGNUM, regs))
		order |= 0x40;
	if (REGSET_CONTAINS_P (HARD_Y_REGNUM, regs))
		order |= 0x20;
	if (REGSET_CONTAINS_P (HARD_X_REGNUM, regs))
		order |= 0x10;
	if (REGSET_CONTAINS_P (HARD_DP_REGNUM, regs))
		order |= 0x8;
	if (REGSET_CONTAINS_P (HARD_B_REGNUM, regs))
		order |= 0x4;
	if (REGSET_CONTAINS_P (HARD_A_REGNUM, regs))
		order |= 0x2;
	if (REGSET_CONTAINS_P (HARD_CC_REGNUM, regs))
		order |= 0x1;

	if (REGSET_CONTAINS_P (HARD_D_REGNUM, regs))
		order |= (0x4 | 0x2);
	return order;
}


/* Returns nonzero if two consecutive push or pop instructions,
 * as determined by the OP, can be merged into a single instruction.
 * The first instruction in the sequence pushes/pops REGS1; the
 * second applies to REGS2.
 *
 * If true, the resulting instruction can use (regs1 | regs2)
 * safely.
 */
int
m6809_can_merge_pushpop_p (int op, int regs1, int regs2)
{
	/* Register sets must not overlap */
	if (regs1 & regs2)
		return 0;

	if (op == UNSPEC_PUSH_RS)
		return (register_push_order (regs1) > register_push_order (regs2));
	else if (op == UNSPEC_POP_RS)
		return (register_push_order (regs1) < register_push_order (regs2));
	else
		return 0;
}


/** Emit instructions for making a library call.
 * MODE is the mode of the operation.
 * NAME is the library function name.
 * OPERANDS is the rtx array provided by the recognizer.
 * COUNT is the number of input operands to the call, and
 * should be 1 for a unary op or 2 for a binary op.
 */
void
emit_libcall_insns (enum machine_mode mode, 
	const char *name, 
	rtx *operands,
	int count)
{
	/* Generate an rtx for the call target. */
	rtx symbol = gen_rtx_SYMBOL_REF (Pmode, name);

	/* Emit the library call.  Slightly different based
	on the number of operands */
	if (count == 2)
		emit_library_call (symbol, 0, mode,
			2, operands[1], mode, operands[2], mode);
	else
		emit_library_call (symbol, 0, mode,
			1, operands[1], mode);

	/* The library call is expected to put its result
	in LIBCALL_VALUE, so need to copy it into the destination. */
	emit_move_insn (operands[0], LIBCALL_VALUE(mode));
}


/**
 * Output a branch/conditional branch insn of the proper
 * length.  code identifies the particular branch insn.
 * operands holds the branch target in operands[0].
 * length says what the size of this insn should be.
 * Based on the length, we know whether it should be a
 * short (8-bit) or long (16-bit) branch.
 */
const char *
output_branch_insn (enum rtx_code code, rtx *operands, int length)
{
	int shortform; 
	const char *opcode;
	char pattern[64];

	/* Decide whether or not to use the long or short form.
	 * Calculate automatically based on insn lengths. */
   shortform = ((length > 2) ? 0 : 1);

	/* Determine the proper opcode.
	 * Use the short (2-byte) opcode if the target is within
	 * reach.  Otherwise, use jmp (3-byte opcode), unless
	 * compiling with -fpic, in which case we'll need to use
	 * lbra (4-byte opcode).
	 */
	switch (code)
	{
		case LABEL_REF: 
			if (shortform)
				opcode = "bra";
			else if (flag_pic)
				opcode = "lbra";
			else
				opcode = "jmp";
			break;
		case EQ: opcode = shortform ? "beq" : "lbeq"; break;
		case NE: opcode = shortform ? "bne" : "lbne"; break;
		case GT: opcode = shortform ? "bgt" : "lbgt"; break;
		case GTU: opcode = shortform ? "bhi" : "lbhi"; break;
		case LT: opcode = shortform ? "blt" : "lblt"; break;
		case LTU: opcode = shortform ? "blo" : "lblo"; break;
		case GE: opcode = shortform ? "bge" : "lbge"; break;
		case GEU: opcode = shortform ? "bhs" : "lbhs"; break;
		case LE: opcode = shortform ? "ble" : "lble"; break;
		case LEU: opcode = shortform ? "bls" : "lbls"; break;
		default: abort(); break;
	}

	/* Build the asm template from the opcode and the target.
	 * Then emit an insn from that template. */
	sprintf (pattern, "%s\t%%l0", opcode);
	output_asm_insn (pattern, operands);
	return "";
}


/** Returns the "cost" of an RTL expression.
 * In general, the expression "COSTS_N_INSNS(1)" is used to represent
 * the cost of a fast 8-bit arithmetic instruction that operates on
 * a reg/mem or a reg/immed.  Other costs are relative to this.
 *
 * Notes:
 * - The cost of a REG is always zero; this cannot be changed.
 *
 * - On the 6809, instructions on two registers will nearly always take
 *   longer than those that operate on a register and a constant/memory,
 *   because of the way the instruction set is structured.
 *
 * TODO: multiply HImode by 2 should be done via shifts, instead of add.
 */
static bool
m6809_rtx_costs (rtx X, int code, int outer_code ATTRIBUTE_UNUSED, int *total)
{
	int has_const_arg = 0;
	HOST_WIDE_INT const_arg;
	enum machine_mode mode;
	int nargs = 1;
	rtx op0, op1;

	/* Data RTXs return a value between 0-3, depending on complexity.
	All of these are less than COSTS_N_INSNS(1). */
	switch (code)
	{
		case CC0:
		case PC:
			*total = 0;
			return true;

 		case CONST_INT:
    		if (X == const0_rtx)
			{
				*total = 0;
				return true;
			}
			else if ((unsigned) INTVAL (X) < 077) 
			{
				*total = 1;
				return true;
			}
			else
			{
				*total = 2;
				return true;
			}

 		case LABEL_REF:
		case CONST:
   		*total = 2;
			return true;

 		case SYMBOL_REF:
			/* References to memory are made cheaper if they have
			the 'direct' mode attribute set. */
			*total = (SYMBOL_REF_FLAG (X)) ? 1 : 2;
			return true;

		case ADDRESSOF:
			*total = COSTS_N_INSNS(1);
			break;

		case MEM:
			/* See what form of address was given */
			X = XEXP (X, 0);
			switch (GET_CODE (X))
			{
 				case SYMBOL_REF:
					*total = (SYMBOL_REF_FLAG (X)) ? 1 : 2;
					break;

				case CONST_INT:
					*total = 2;
					break;

				case MEM:
					*total = COSTS_N_INSNS (1) + 2;
					break;

				default:
					break;
			}
			return true;

 		case CONST_DOUBLE:
   		*total = 3;
			return true;

		default:
			break;
	}

	/* Decode the rtx */
	mode = GET_MODE (X);
	op0 = XEXP (X, 0);
	op1 = XEXP (X, 1);

	/* We don't implement anything in SImode or greater. */
	if (GET_MODE_SIZE (mode) >= GET_MODE_SIZE (SImode))
	{
		*total = COSTS_N_INSNS (100);
		return true;
	}

	/* Figure out if there is a constant argument, and its value. */
#if (TARGET_GCC_VERSION < 4000)
	if ((GET_RTX_CLASS (code) == '2') || (GET_RTX_CLASS (code) == 'c'))
#else
	if (GET_RTX_CLASS (code) == RTX_BIN_ARITH
      || GET_RTX_CLASS (code) == RTX_COMM_ARITH)
#endif
	{
		nargs = 2;
		if (GET_CODE (op1) == CONST_INT)
		{
			has_const_arg = 1;
			const_arg = INTVAL (op1);
		}
	}

	/* Penalize a reg/reg operation by adding MEMORY_MOVE_COST,
	 * Ignore soft registers, since these are really in memory. */
	if (nargs == 2 && REAL_REG_P (op0) && REAL_REG_P (op1))
	{
		*total = MEMORY_MOVE_COST (mode, Q_REGS, 0);
	}
	else
	{
		*total = 0;
	}

	/* Operator RTXs return a multiple of COSTS_N_INSNS(1). */
	switch (code)
	{
		case PLUS: case MINUS: case COMPARE:
			/* 6809 handles these natively in QImode, and in HImode as long
			 * as operand 1 is constant. */
			if (mode == QImode || (mode == HImode && has_const_arg))
				*total += COSTS_N_INSNS (1);
			else 
				*total += COSTS_N_INSNS (GET_MODE_SIZE (mode));

			/* -1, 0, and 1 can be done using inherent instructions
			 * for PLUS and MINUS, so don't add extra cost. */
  			if (has_const_arg
				&& (const_arg == -1 || const_arg == 0 || const_arg == 1)
				&& (code == PLUS || code == MINUS))
			{
				return true;
			}
			break;

		case AND: case IOR: case XOR:
		case NEG: case NOT:
			/* 6809 handles these natively in QImode, but requires
			 * splitting in HImode.   Treat these as 2 insns. */
			*total += COSTS_N_INSNS (1) * GET_MODE_SIZE (mode);
			break;

  		case ASHIFT: case ASHIFTRT: case LSHIFTRT:
  		case ROTATE: case ROTATERT:
			/* 6809 can do shift/rotates of a QImode by a constant in
			 * 1 insn times the shift count, or in HImode by a constant 
			 * by splitting to 2 insns.
			 *
			 * Shift by a nonconstant will take significantly longer
			 * than any of these. */
  			if (has_const_arg)
			{
				/* Truncate the shift count to the mode size */
				const_arg %= (GET_MODE_SIZE (mode) * 8);

				/* HImode shifts greater than 8 get optimized due
				 * to register transfer from b to a; this cuts down the
				 * cost. */
				if (const_arg >= 8)
				{
					*total += COSTS_N_INSNS (1);
					const_arg -= 8;
				}

				/* The computed cost is 'const_arg' 1-bit shifts, doubled
				if in HImode, minus the cost of the constant itself which
				will be added in later but really shouldn't be. */
				*total += COSTS_N_INSNS (const_arg) * GET_MODE_SIZE (mode) - 1;
				return true;
			}
			else
			{
				/* It may take up to 7 iterations of about 6-7 real
				 * instructions, so make this expensive. */
				*total += COSTS_N_INSNS (50);
			}
  			break;

		case MULT:
		{
			/* Multiply is cheap when both arguments are 8-bits.  They
			could be QImode, or QImode widened to HImode, or a constant
			that fits into 8-bits.  As long as both operands qualify,
			we can use a single mul instruction.

			Assume that fast multiply can be used, and change this if we find
			differently... */
			int ok_for_qihi3 = 1;

			/* Check the first operand */	
			switch (GET_MODE (op0))
			{
				case QImode:
					break;
				case HImode:
					if (GET_CODE (op0) != SIGN_EXTEND && GET_CODE (op0) != ZERO_EXTEND)
						ok_for_qihi3 = 0;
					break;
				default:
					ok_for_qihi3 = 0;
					break;
			}

			/* Likewise, check the first operand.  This is where constants may appear. */	
			switch (GET_MODE (op1))
			{
				case QImode:
					break;
				case HImode:
					if (GET_CODE (op0) != SIGN_EXTEND && GET_CODE (op0) != ZERO_EXTEND)
						ok_for_qihi3 = 0;
					break;
				case VOIDmode:
				if (!CONST_OK_FOR_LETTER_P (const_arg, 'K'))
						ok_for_qihi3 = 0;
				default:
					ok_for_qihi3 = 0;
					break;
			}

			/* Fast multiply takes about 4 times as many cycles as a normal
			arithmetic operation.  Otherwise, it will take an expensive libcall. */
			if (ok_for_qihi3)
				*total += COSTS_N_INSNS (4);
			else
				*total = COSTS_N_INSNS (50);
	  		break;
		}

		case DIV: case UDIV: case MOD: case UMOD:
			/* These all require more expensive libcalls */
			*total += COSTS_N_INSNS (100);
  			break;

		/* TODO : TRUNCATE, SIGN_EXTEND, and ZERO_EXTEND */

		/* These can normally be done with autoincrement, etc., so
		 * don't charge for them. */
		case PRE_DEC:
		case PRE_INC:
		case POST_DEC:
		case POST_INC:
			*total = 0;
			break;

		default:
			break;
	}

	/* Always return false, and let the caller gather the costs
	 * of the operands */
	return false;
}


static tree
m6809_handle_fntype_attribute (tree *node, tree name,
	tree args ATTRIBUTE_UNUSED,
	int flags ATTRIBUTE_UNUSED,
	bool *no_add_attrs)
{
	if (TREE_CODE (*node) != FUNCTION_TYPE)
	{
		warning ("'%s' only valid for functions", 
			IDENTIFIER_POINTER (name));
		*no_add_attrs = TRUE;
	}

	return NULL_TREE;
}


static tree
m6809_handle_data_type_attribute (tree *node, tree name,
	tree args ATTRIBUTE_UNUSED,
	int flags ATTRIBUTE_UNUSED,
	bool *no_add_attrs)
{
	return NULL_TREE;
}



static tree
m6809_handle_default_attribute (tree *node ATTRIBUTE_UNUSED, 
	tree name ATTRIBUTE_UNUSED,
	tree args ATTRIBUTE_UNUSED,
	int flags ATTRIBUTE_UNUSED,
	bool *no_add_attrs ATTRIBUTE_UNUSED )
{
	return NULL_TREE;
}


/* Table of valid machine attributes */
const struct attribute_spec m6809_attribute_table[] = { /*
{ name,        min, max, decl,  type, fntype, handler } */
{ "interrupt", 0,   0,   false, true,  true,  m6809_handle_fntype_attribute },
{ "naked",     0,   0,   false, true,  true,  m6809_handle_fntype_attribute },
{ "far",       0,   1,   true,  false, false, m6809_handle_default_attribute },
{ "bank",      0,   1,   true,  false, false, m6809_handle_default_attribute },
{ "boolean",   0,   0,   false, true,  false, m6809_handle_data_type_attribute },
{ NULL,        0,   0,   false, true,  false, NULL },
};


/** Initialize builtin routines for the 6809. */
void
m6809_init_builtins (void)
{
	/* Create type trees for each function signature required.
	 *
	 * void_ftype_void = void f(void)
	 * void_ftype_uchar = void f(unsigned char)
	 * uchar_ftype_uchar2 = unsigned char f (unsigned char, unsigned char)
	 */
	tree void_ftype_void = 
		build_function_type (void_type_node, void_list_node);

	tree void_ftype_uchar =
		build_function_type (void_type_node,
			tree_cons (NULL_TREE, unsigned_char_type_node, void_list_node));

	tree uchar_ftype_uchar2 =
		build_function_type (unsigned_char_type_node,
			tree_cons (NULL_TREE, unsigned_char_type_node, 
				tree_cons (NULL_TREE, unsigned_char_type_node, void_list_node)));

	/* Register each builtin function. */
	builtin_function ("__builtin_swi", void_ftype_void,
		M6809_SWI, BUILT_IN_MD, NULL, NULL_TREE);

	builtin_function ("__builtin_swi2", void_ftype_void,
		M6809_SWI2, BUILT_IN_MD, NULL, NULL_TREE);

	builtin_function ("__builtin_swi3", void_ftype_void,
		M6809_SWI3, BUILT_IN_MD, NULL, NULL_TREE);

	builtin_function ("__builtin_cwai", void_ftype_uchar,
		M6809_CWAI, BUILT_IN_MD, NULL, NULL_TREE);

	builtin_function ("__builtin_sync", void_ftype_void,
		M6809_SYNC, BUILT_IN_MD, NULL, NULL_TREE);

	builtin_function ("__builtin_nop", void_ftype_void,
		M6809_NOP, BUILT_IN_MD, NULL, NULL_TREE);

	builtin_function ("__builtin_blockage", void_ftype_void,
		M6809_BLOCKAGE, BUILT_IN_MD, NULL, NULL_TREE);

	builtin_function ("__builtin_add_decimal", uchar_ftype_uchar2,
		M6809_ADD_DECIMAL, BUILT_IN_MD, NULL, NULL_TREE);

	builtin_function ("__builtin_add_carry", uchar_ftype_uchar2,
		M6809_ADD_CARRY, BUILT_IN_MD, NULL, NULL_TREE);

	builtin_function ("__builtin_sub_carry", uchar_ftype_uchar2,
		M6809_SUB_CARRY, BUILT_IN_MD, NULL, NULL_TREE);
}


/** Used by m6809_expand_builtin, given a tree ARGLIST which
 * refers to the operands of a builtin call, return an rtx
 * that represents the nth operand, as denoted by OPNUM, which
 * is a zero-based integer.  MODE gives the expected mode
 * of the operand.
 *
 * This rtx is suitable for use in the emitted RTL for the
 * builtin instruction. */
rtx
m6809_builtin_operand (tree arglist, enum machine_mode mode, int opnum)
{
	tree arg;
	rtx r;

	/* First get the tree for the argument */
	switch (opnum)
	{
		case 0:
			arg = TREE_VALUE (arglist);
			break;

		case 1:
			arg = TREE_VALUE (TREE_CHAIN (arglist));
			break;

		case 2:
			arg = TREE_VALUE (TREE_CHAIN (TREE_VALUE (TREE_CHAIN (arglist))));
			break;

		default:
			return NULL_RTX;
	}

	/* Convert the tree to RTL */
	r = expand_expr (arg, NULL_RTX, mode, 0);
	if (r == NULL_RTX)
		return NULL_RTX;

	/* Not sure why this is needed, but every other backend does it. */
	r = protect_from_queue (r, 0);
	if (r == NULL_RTX)
		return NULL_RTX;

	return r;
}


/** Expand a builtin that was registered in init_builtins into
 * RTL.  */
rtx
m6809_expand_builtin (tree exp, 
	rtx target, 
	rtx subtarget ATTRIBUTE_UNUSED,
	enum machine_mode mode ATTRIBUTE_UNUSED,
	int ignore ATTRIBUTE_UNUSED )
{
	tree fndecl = TREE_OPERAND (TREE_OPERAND (exp, 0), 0);
	unsigned int fcode = DECL_FUNCTION_CODE (fndecl);
	tree arglist = TREE_OPERAND (exp, 1);
	rtx r0, r1;

	switch (fcode)
	{
		case M6809_SWI:
			r0 = gen_rtx_CONST_INT (VOIDmode, 1);
			emit_insn (target = gen_m6809_swi (r0));
			return target;

		case M6809_SWI2:
			r0 = gen_rtx_CONST_INT (VOIDmode, 2);
			emit_insn (target = gen_m6809_swi (r0));
			return target;

		case M6809_SWI3:
			r0 = gen_rtx_CONST_INT (VOIDmode, 3);
			emit_insn (target = gen_m6809_swi (r0));
			return target;

		case M6809_CWAI:
			r0 = m6809_builtin_operand (arglist, QImode, 0);
			emit_insn (target = gen_m6809_cwai (r0));
			return target;

		case M6809_SYNC:
			emit_insn (target = gen_m6809_sync ());
			return target;

		case M6809_ADD_CARRY:
			r0 = m6809_builtin_operand (arglist, QImode, 0);
			r1 = m6809_builtin_operand (arglist, QImode, 1);
			if (!target)
				target = gen_reg_rtx (QImode);
			emit_insn (gen_addqi3_carry (target, r0, r1));
			return target;

		case M6809_SUB_CARRY:
			r0 = m6809_builtin_operand (arglist, QImode, 0);
			r1 = m6809_builtin_operand (arglist, QImode, 1);
			if (!target)
				target = gen_reg_rtx (QImode);
			emit_insn (gen_subqi3_carry (target, r0, r1));
			return target;

		case M6809_NOP:
			emit_insn (target = gen_nop ());
			return target;

		case M6809_BLOCKAGE:
			emit_insn (target = gen_blockage ());
			return target;

		case M6809_ADD_DECIMAL:
			r0 = m6809_builtin_operand (arglist, QImode, 0);
			r1 = m6809_builtin_operand (arglist, QImode, 1);
			if (!target)
				target = gen_reg_rtx (QImode);
			emit_insn (gen_addqi3_decimal (target, r0, r1));
			return target;

		default:
			warning ("unknown builtin expansion ignored");
			return NULL_RTX;
	}
}



/* Returns nonzero if 'x' represents a function that was declared
 * as __noreturn__. */
int
noreturn_functionp (rtx x)
{
	tree decl = call_target_decl (x);

	if (decl == NULL_TREE)
		return 0;
	else
		return TREE_THIS_VOLATILE (decl);
}


/* For a far function, returns the identifier that states which page
 * it resides in.  Otherwise, returns NULL for ordinary functions. */
const char *
far_functionp (rtx x)
{
	tree decl, attr;
	const char *page;

	decl = call_target_decl (x);
	if (decl == NULL_TREE)
		return NULL;

	/* See if the function has the new 'banked' attribute.  These
	 * are numeric instead of text */
	page = m6809_get_decl_bank (decl);
	if (page)
		return page;

	/* Return whether or not this decl has the far attribute */
	attr = lookup_attribute ("far", DECL_ATTRIBUTES (decl));
	if (attr == NULL_TREE)
		return NULL;

	/* If it is far, check for a value */
	attr = TREE_VALUE (attr);
	if (attr == NULL_TREE)
	{
		warning ("far code page not specified, using local value");
		return far_code_page;
	}

	/* We have a TREE_LIST of attribute values, get the first one.
	 * It should be an INTEGER_CST. */
	attr = TREE_VALUE (attr);
	page = TREE_STRING_POINTER (attr);
	return page;
}


/** Outputs the assembly language for a far call. */
void
output_far_call_insn (rtx *operands, int has_return)
{
	static char page_data[64];
	const char *called_page;

	/* The logic is the same for functions whether or not there
	 * is a return value.  Skip over the return value in this
	 * case, so that the call location is always operands[0].  */
	if (has_return)
		operands++;

  /* Get the name of the page being called */
  called_page = far_functionp (operands[0]);

#if 0 /* TODO : broken logic */
  /* See if the called page name is a 'bank' */
  if (isdigit (*called_page))
  {
    /* New style banking */
	 if (!strcmp (called_page, current_bank_name))
	 {
	 	/* Same page */
  	  	output_asm_insn ("jsr\t%0", operands);
	 }
	 else
	 {
	 	/* Different page */
		output_asm_insn ("jsr\t__far_call_handler\t;new style", operands);
  	  	output_asm_insn ("\t.dw\t%0", operands);
		sprintf (page_data, "\t.db\t%s", called_page);
	 	output_asm_insn (page_data, operands);
	 }
	 return;
  }
#endif

  /* Are we calling a different page than we are running in? */
  if (!strcmp (called_page, far_code_page))
  {
    /* Same page : no need to execute a far call */
		if (flag_pic)
			output_asm_insn ("lbsr\t%C0", operands);
		else
			output_asm_insn ("jsr\t%0", operands);
  }
  else
  {
    /* Different page : need to emit far call thunk */

    /* First output a call to the thunk for making far calls. */
		if (flag_pic)
			output_asm_insn ("lbsr\t__far_call_handler", operands);
		else
			output_asm_insn ("jsr\t__far_call_handler\t;old style", operands);
  
    /* Now output the name of the call site */
    output_asm_insn ("\t.dw\t%C0", operands);
  
    /* Finally output the page number */
    sprintf (page_data, "\t.db\t%s", far_functionp (operands[0]));
    output_asm_insn (page_data, operands);
  }
}


/*
 * Trampoline output:
 *
 * ldu #&cxt      4 bytes   --LDY- ?? ??
 * jmp fnaddr     3 bytes   JMP ?? ??
 */
void
m6809_initialize_trampoline (rtx tramp, rtx fnaddr ATTRIBUTE_UNUSED, rtx cxt)
{
	/* TODO - optimize by generating the entire trampoline code here,
	 * and removing the template altogether, since there are only two
	 * bytes there that matter. */
	emit_move_insn (gen_rtx_MEM (HImode, plus_constant (tramp, 2)), cxt);
	emit_move_insn (gen_rtx_MEM (HImode, plus_constant (tramp, 5)), fnaddr);
}


/** Echo the version of the compiler and the name of the source file
 * at the beginning of each assembler output file.  asm_out_file
 * is a global FILE * pointing to the output stream. */
void
m6809_asm_file_start (void)
{
	char *module_name;
	
	fprintf (asm_out_file, "\n;;; gcc for m6809 : %s %s %s\n",
		__DATE__, __TIME__, subversion_tag);

	fprintf (asm_out_file, ";;; ABI version %d\n", m6809_abi_version);
	fprintf (asm_out_file, ";;; %s\n",
		(TARGET_BYTE_INT ? "-mint8" : "-mint16"));
	if (TARGET_EXPERIMENT)
		fprintf (asm_out_file, ";;; -mexperiment\n");
	if (TARGET_WPC)
		fprintf (asm_out_file, ";;; -mwpc\n");
	if (TARGET_6309)
		fprintf (asm_out_file, ";;; -m6309\n");

	module_name = my_basename (main_input_filename);
	fprintf (asm_out_file, "\t.module\t%s%s\n", 
		isalpha (*module_name) ? "" : "_",
		module_name);
}


/** Returns true if prologue/epilogue code is required for the
 * current function being compiled.
 *
 * This is just the inverse of whether the function is declared as
 * 'naked'.
 */
int
prologue_epilogue_required (void)
{
	return !m6809_current_function_has_type_attr_p ("naked")
		&& !m6809_current_function_has_type_attr_p ("noreturn");
}


/** Expand RTL for function entry */
void
emit_prologue_insns (void)
{
  unsigned int live_regs = m6809_get_live_regs ();
  unsigned int frame_size = m6809_get_actual_frame_size (false);

  if (live_regs && !m6809_current_function_has_type_attr_p ("interrupt"))
    emit_insn (
      gen_rtx_register_pushpop (UNSPEC_PUSH_RS, live_regs));

  if (frame_size != 0)
    emit_insn (gen_rtx_stack_adjust (MINUS, frame_size));
 
  /* Establish a new frame if necessary */
  if (frame_pointer_needed)
    emit_move_insn (hard_frame_pointer_rtx, stack_pointer_rtx);
}


/** Expand RTL for function exit */
void
emit_epilogue_insns (void)
{
  unsigned int live_regs = m6809_get_live_regs ();
  unsigned int frame_size = m6809_get_actual_frame_size (false);

  if (frame_size != 0)
    emit_insn (gen_rtx_stack_adjust (PLUS, frame_size));
 
  if (live_regs && !m6809_current_function_has_type_attr_p ("interrupt"))
      emit_insn (
        gen_rtx_register_pushpop (UNSPEC_POP_RS, PC_REGBIT | live_regs));

  if (m6809_current_function_has_type_attr_p ("interrupt"))
    emit_insn (gen_return_from_interrupt_epilogue ());
  else if (live_regs == 0)
    emit_insn (gen_return_from_epilogue ());
}


/** Predefine some preprocessor names according to the currently
 * selected compiler options */
void
m6809_cpu_cpp_builtins (void)
{
	if (TARGET_6309)
	{
		builtin_define_std ("__M6309__");
		builtin_define_std ("__m6309__");
	}
	else
	{
		builtin_define_std ("__M6809__");
		builtin_define_std ("__m6809__");
	}

	if (TARGET_BYTE_INT)
		builtin_define_std ("__int8__");
	else
		builtin_define_std ("__int16__");

	switch (m6809_abi_version)
	{
		case M6809_ABI_VERSION_STACK:
			builtin_define_std ("__regargs__");
			builtin_define_std ("__ABI_STACK__");
			break;
		case M6809_ABI_VERSION_REGS:
			builtin_define_std ("__ABI_REGS__");
			break;
		case M6809_ABI_VERSION_BX:
			builtin_define_std ("__ABI_BX__");
			break;
		default:
			break;
	}

	if (TARGET_WPC)
		builtin_define_std ("__WPC__");
}


#define MAX_ASM_ASCII_STRING 48

void
m6809_output_ascii (FILE *fp, const char *str, unsigned long size)
{
	unsigned long i;

	/* If the size is too large, then break this up into multiple
	outputs.  The assembler can only output roughly 48 bytes at a
	time.  Note that if there are lots of escape sequences in
	the string, this may fail. */
	if (size > MAX_ASM_ASCII_STRING)
	{
		m6809_output_ascii (fp, str, MAX_ASM_ASCII_STRING);
		m6809_output_ascii (fp, str + MAX_ASM_ASCII_STRING, 
			size - MAX_ASM_ASCII_STRING);
		return;
	}

	fprintf (fp, "\t.ascii \"");

	for (i = 0; i < size; i++)
	{
		int c = str[i] & 0377;

		/* Just output the plain character if it is printable,
		otherwise output the escape code for the character.
		The assembler recognizes the same C-style escape sequences. */
		if (c >= ' ' && c < 0177 && c != '\\' && c != '"')
  			putc (c, fp);
		else switch (c) 
		{
			case '\r':
				fputs ("\\r", fp);
				break;
			case '\n':
				fputs ("\\n", fp);
				break;
			case '\t':
				fputs ("\\t", fp);
				break;
			case '\f':
				fputs ("\\f", fp);
				break;
			case 0:
				fputs ("\\0", fp);
				break;
			default:
				fprintf (fp, "\\%03o", c);
				break;
		}
	}
	fprintf (fp, "\"\n");
}


void
m6809_output_quoted_string (FILE *asm_file, const char *string)
{
	char c;

	if (strlen (string) > MAX_ASM_ASCII_STRING)
	{
		/* The string length is too large.  We'll have to truncate it.
		This is only called from debugging functions, so it's usually
		not critical. */

		char truncated_string[MAX_ASM_ASCII_STRING+1];

		/* Copy as many characters as we can. */
		strncpy (truncated_string, string, MAX_ASM_ASCII_STRING);
		truncated_string[MAX_ASM_ASCII_STRING] = '\0';
		string = truncated_string;
	}

	/* Copied from toplev.c */

	putc ('\"', asm_file);
	while ((c = *string++) != 0) {
		if (ISPRINT (c)) {
			if (c == '\"' || c == '\\')
				putc ('\\', asm_file);
			putc (c, asm_file);
		}
      else
			fprintf (asm_file, "\\%03o", (unsigned char) c);
	}
	putc ('\"', asm_file);
}


/** A table of shift opcodes and their properties, used for generating
optimal sequences of assembly for shifts/rotates by constant values. */
struct shift_opcode
{
	/* The mnemonic to use when performing the operation on a MEM */
	const char *asm_mem;

	/* The mnemonic to use when performing the operation on the
	high byte of a 16-bit quantity. */
	const char *asm_high;

	/* The mnemonic to use when performing the operation on the
	low byte of a 16-bit quantity. */
	const char *asm_low;

	/* A flag that says whether this is a shift or a rotate */
	enum { SOP_TYPE_SHIFT, SOP_TYPE_ROTATE } type;

	/* A flag that says whether this op moves to the left or right */
	enum { SOP_DIR_LEFT, SOP_DIR_RIGHT } dir;

	/* A pointer to the opcode info for an 'extend' operation. */
	struct shift_opcode *extend;

	/* A pointer to the opcode info for the negation of this operation */
	struct shift_opcode *inverse;
};


static struct shift_opcode shift_opcode_table[] = {
	[ASHIFT] = { "asl", "asla", "aslb", SOP_TYPE_SHIFT, SOP_DIR_LEFT, 
		&shift_opcode_table[ROTATE], &shift_opcode_table[ASHIFTRT] },
	[ASHIFTRT] = { "asr", "asra", "asrb", SOP_TYPE_SHIFT, SOP_DIR_RIGHT, 
		&shift_opcode_table[ROTATERT], &shift_opcode_table[ASHIFT] },
	[LSHIFTRT] = { "lsr", "lsra", "lsrb", SOP_TYPE_SHIFT, SOP_DIR_RIGHT, 
		&shift_opcode_table[ROTATERT], &shift_opcode_table[ASHIFT] },
	[ROTATE] = { "rol", "rola", "rolb", SOP_TYPE_ROTATE, SOP_DIR_LEFT, 
		&shift_opcode_table[ROTATE], &shift_opcode_table[ROTATERT] },
	[ROTATERT] = { "ror", "rora", "rorb", SOP_TYPE_ROTATE, SOP_DIR_RIGHT, 
		&shift_opcode_table[ROTATERT], &shift_opcode_table[ROTATE] },
};


/* Expand a shift by a constant into a series of instructions. */
void
expand_constant_shift (int code, rtx dst, rtx src, rtx count)
{
	int n = INTVAL (count);
	enum machine_mode mode = GET_MODE (dst);
	rtx mask = NULL_RTX;

	rtx areg_rtx = gen_rtx_REG (QImode, HARD_A_REGNUM);
	rtx breg_rtx = gen_rtx_REG (QImode, HARD_D_REGNUM);

	/* TODO : Need logic to decide whether a shift of a MEM
	is best done in-place, or by preloading it into a register
	and shifting that.  This should be covered by rtx_costs. */

	/* Sanity check */
	if (code != ASHIFT && code != ASHIFTRT && code != LSHIFTRT)
		abort ();

	/* When shifting a 16-bit value, which is guaranteed to be
	in 'D' register because of constraints, the first 8 bits
	of the shift can be done fast by shifting values between
	the 'A' and 'B' registers. */
	if (n >= 8 && mode == HImode)
	{
		if (code == ASHIFT)
		{
			/* Emit tfr b,a; clrb */
			emit_move_insn (areg_rtx, breg_rtx);
			emit_insn (gen_rtx_USE (VOIDmode, areg_rtx));
			emit_move_insn (breg_rtx, const0_rtx);

			/* Operate on 'A' from now on */
			dst = src = areg_rtx;
		}
		else
		{
			/* Emit tfr a,b; clra */
			emit_move_insn (breg_rtx, areg_rtx);
			emit_move_insn (areg_rtx, const0_rtx);
			emit_insn (gen_rtx_USE (VOIDmode, areg_rtx));

			/* Operate on 'B' from now on */
			dst = src = breg_rtx;
		}

		mode = QImode;
		n -= 8;
	}

	/* Shifts of 6 or 7 bits are done faster by rotating in the
	opposite direction, then masking the result. */
	if (mode == QImode && n >= 6)
	{
		/* Calculate the amount to mask the result */
		if (code == ASHIFT)
			mask = gen_rtx_CONST_INT (VOIDmode, (n == 6) ? 0xC0 : 0x80);
		else
			mask = gen_rtx_CONST_INT (VOIDmode, (n == 6) ? 0x03 : 0x01);

		/* Convert the shift operator into a rotation.
		These are 6809 rotations, which go through the carry flag.
		Thus the extra 1 in the adjusted count. */
		if (code == ASHIFT)
			code = ROTATERT;
		else
			code = ROTATE;
		n = GET_MODE_BITSIZE (mode) + 1 - n;
	}

	/* Unroll the shift/rotation, one instruction at a time. */
	while (n-- > 0)
	{
		emit_move_insn (dst, gen_rtx (code, mode, dst, src, const1_rtx));
		if (mode == HImode)
		{
		}
	}

	/* Apply a mask if necessary */
	if (mask != NULL_RTX)
	{
		emit_move_insn (dst, gen_rtx (AND, mode, dst, src, mask));
	}
}



/** The old way to output assembler code for a shift or rotate expression. */
void
m6809_output_shift_insn (int rtx_code, rtx *operands)
{
	struct shift_opcode *op;
	int c;
	enum machine_mode mode;
	int memp;
	rtx xoperands[4];
	rtx dest = operands[0];
	rtx src = operands[1];
	rtx count = operands[2];
	char asm_string[32];
	int mem_was_high = 0;
	int sign_extend_needed = 0;

	for (c = 0; c < 3; c++)
		xoperands[c] = operands[c];
	xoperands[3] = NULL;

	/* See what type of shift/rotate is being requested */
	op = &shift_opcode_table[rtx_code];
	if (!op)
		abort ();

	/* If the shift count is not constant, there's already a function
	for handling that. */
	if (GET_CODE (count) != CONST_INT)
	{
		m6809_gen_register_shift (xoperands, op->asm_low, op->extend->asm_high);
		return;
	}

	/* Check various attributes of the insn. */
	mode = GET_MODE (dest);
	memp = MEM_P (dest);
	c = INTVAL (count);

	/* Sanity checks */
	if (op->dir != op->extend->dir)
		abort ();
	if (op->dir == op->inverse->dir)
		abort ();
	if (mode != GET_MODE (src))
		abort ();
	if (!memp && GET_CODE (src) != REG)
		abort ();
	if ((mode != QImode) && (mode != HImode))
		abort ();

	/* Check for various things that can be optimized. */

	/* Ensure the rotate count is modulo the mode size.
	 * Also, force rotate through the carry bit back into the data
	 * by increasing the size by 1. */
	if (op->type == SOP_TYPE_ROTATE)
	{
		c = (c + 1) % GET_MODE_BITSIZE (mode);

		/* Rotates by a large number can be simplified by rotating 
		in the opposite direction. */
		if (c > GET_MODE_BITSIZE (mode) - 3)
		{
			c = (GET_MODE_BITSIZE (mode) + 1) - c;
			op = op->inverse;
		}
	}

	/* Shifts >= 8 can be factored into a shift by 8, followed by a 
	shift of N-8.  Take care not to lose the sign bit for an arithmetic
	right shift. */
	if ((op->type == SOP_TYPE_SHIFT) && (c >= 8))
	{
		if (mode != HImode)
			abort ();

		if (memp)
		{
			/* A mem-mem copy requires a scratch reg.  Just use B for now. */
			output_asm_insn ("pshs\tb", xoperands);
			if (op->dir == SOP_DIR_LEFT)
			{
				output_asm_insn ("ldb\t%L0", xoperands);
				output_asm_insn ("stb\t%0", xoperands);
				output_asm_insn ("clr\t%L0", xoperands);
			}
			else
			{
				output_asm_insn ("ldb\t%0", xoperands);
				output_asm_insn ("stb\t%L0", xoperands);
				if (rtx_code == ASHIFTRT)
				{
					output_asm_insn ("rolb", xoperands);
					output_asm_insn ("rolb", xoperands);
					output_asm_insn ("andb\t#1", xoperands);
					output_asm_insn ("negb", xoperands);
					output_asm_insn ("stb\t%0", xoperands);
				}
				else
				{
					output_asm_insn ("clr\t%0", xoperands);
				}
				mem_was_high = 1;
			}
			output_asm_insn ("puls\tb", xoperands);
		}
		else
		{
			if (op->dir == SOP_DIR_LEFT)
			{
				output_asm_insn ("tfr\tb,a", xoperands);
				output_asm_insn ("clrb", xoperands);
				xoperands[0] = xoperands[1] = gen_rtx_REG (QImode, HARD_A_REGNUM);
			}
			else
			{
				output_asm_insn ("tfr\ta,b", xoperands);
				if (rtx_code == ASHIFTRT)
				{
					output_asm_insn ("rola", xoperands);
					output_asm_insn ("rola", xoperands);
					output_asm_insn ("anda\t#1", xoperands);
					output_asm_insn ("nega", xoperands);
				}
				else
				{
					output_asm_insn ("clra", xoperands);
				}
				xoperands[0] = xoperands[1] = gen_rtx_REG (QImode, HARD_D_REGNUM);
			}
		}

		mode = QImode;
		c -= 8;
	}

	/* Unsigned 8-bit shifts by a large number can be simplified by rotating
	in the opposite direction, and then masking off the bits guaranteed
	to be zero. */
	if (op->type == SOP_TYPE_SHIFT && rtx_code != ASHIFTRT
		&& (mode == QImode) && (c >= 6) && !memp)
	{
		if (op->dir == SOP_DIR_LEFT)
			xoperands[3] = gen_rtx_CONST_INT (VOIDmode, (c == 6) ? 0xC0 : 0x80);
		else
			xoperands[3] = gen_rtx_CONST_INT (VOIDmode, (c == 6) ? 0x03 : 0x01);
		op = op->extend->inverse;
		c = GET_MODE_BITSIZE (mode) + 1 - c;
	}

	/* Signed 8-bit right shift by 7 is a special case */
	if (rtx_code == ASHIFTRT && c == 7)
	{
		/* Note: if this was originally a HImode shift by 15,
		 * then this is not optimal. */
		output_asm_insn ("rolb", xoperands);
		output_asm_insn ("rolb", xoperands);
		output_asm_insn ("andb\t#1", xoperands);
		output_asm_insn ("negb", xoperands);
		return;
	}

	/* Default case is to expand the opcode into C instructions.
	Note that c might be zero. */
	while (c-- > 0)
	{
		if (mode == QImode)
		{
			if (memp)
			{
				if (mem_was_high)
					sprintf (asm_string, "%s\t%%L0", op->asm_mem);
				else
					sprintf (asm_string, "%s\t%%0", op->asm_mem);
				output_asm_insn (asm_string, xoperands);
			}
			else
			{
				if (REGNO (xoperands[0]) == HARD_A_REGNUM)
					output_asm_insn (op->asm_high, xoperands);
				else
					output_asm_insn (op->asm_low, xoperands);
			}
		}
		else if (mode == HImode)
		{
			if (memp)
			{
				if (op->dir == SOP_DIR_LEFT)
				{
					sprintf (asm_string, "%s\t%%L0", op->asm_mem);
					output_asm_insn (asm_string, xoperands);
					sprintf (asm_string, "%s\t%%0", op->extend->asm_mem);
					output_asm_insn (asm_string, xoperands);
				}
				else
				{
					sprintf (asm_string, "%s\t%%0", op->extend->asm_mem);
					output_asm_insn (asm_string, xoperands);
					sprintf (asm_string, "%s\t%%L0", op->asm_mem);
					output_asm_insn (asm_string, xoperands);
				}
			}
			else
			{
				if (op->dir == SOP_DIR_LEFT)
				{
					output_asm_insn (op->asm_low, xoperands);
					output_asm_insn (op->extend->asm_high, xoperands);
				}
				else
				{
					output_asm_insn (op->asm_high, xoperands);
					output_asm_insn (op->extend->asm_low, xoperands);
				}
			}
		}
		else
			abort ();
	}

	if (xoperands[3] != NULL)
	{
		output_asm_insn ("and%0\t%3", xoperands);
	}
}


/** Adjust register usage based on compile-time flags. */
void
m6809_conditional_register_usage (void)
{
	unsigned int soft_regno;

	if (!m6809_soft_reg_count)
		return;

	m6809_soft_regs = atoi (m6809_soft_reg_count);
	if (m6809_soft_regs == 0)
		return;

	if (m6809_soft_regs > NUM_M_REGS)
		m6809_soft_regs = NUM_M_REGS;

	/* Registers are marked FIXED by default.  Free up if
	the user wishes. */
	for (soft_regno = 0; soft_regno < m6809_soft_regs; soft_regno++)
	{
		fixed_regs[SOFT_M0_REGNUM + soft_regno] = 0;

		/* Mark the softregs as call-clobbered, so that they need
		 * not be saved/restored on function entry/exit. */
		call_used_regs[SOFT_M0_REGNUM + soft_regno] = 1;
	}
}


rtx
m6809_function_value (tree valtype, tree func)
{
   unsigned int regno;
	enum machine_mode mode;

	mode = TYPE_MODE (valtype);

	/* Promote QImode to HImode */
	/* TODO - ?this still causes some problems while fixing others */
//	if (!TARGET_BYTE_INT && mode == QImode)
//		mode = HImode;

	if (lookup_attribute ("boolean", TYPE_ATTRIBUTES (valtype)))
      regno = HARD_Z_REGNUM;
   else if (mode == QImode)
      regno = HARD_D_REGNUM;
   else
      regno = HARD_X_REGNUM;
   return gen_rtx_REG (mode, regno);
}


int
m6809_function_value_regno_p (unsigned int regno)
{
	if (regno == HARD_Z_REGNUM)
		return 1;
	else if (TARGET_BYTE_INT && regno == HARD_D_REGNUM)
		return 1;
	else if (regno == HARD_X_REGNUM)
		return 1;
	else
		return 0;
}


int
power_of_two_p (unsigned int n)
{
	return (n & (n-1)) == 0;
}


/* Defines the target-specific hooks structure. */
struct gcc_target targetm = TARGET_INITIALIZER;

